/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.h
  * @brief          : Header for main.c file.
  *                   This file contains the common defines of the application.
  ******************************************************************************
  * @attention
  *
  * Copyright (c) 2025 STMicroelectronics.
  * All rights reserved.
  *
  * This software is licensed under terms that can be found in the LICENSE file
  * in the root directory of this software component.
  * If no LICENSE file comes with this software, it is provided AS-IS.
  *
  ******************************************************************************
  */
/* USER CODE END Header */

/* Define to prevent recursive inclusion -------------------------------------*/
#ifndef __MAIN_H
#define __MAIN_H

#ifdef __cplusplus
extern "C" {
#endif

/* Includes ------------------------------------------------------------------*/
#include "stm32f4xx_hal.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */

/* USER CODE END Includes */

/* Exported types ------------------------------------------------------------*/
/* USER CODE BEGIN ET */

/* USER CODE END ET */

/* Exported constants --------------------------------------------------------*/
/* USER CODE BEGIN EC */

/* USER CODE END EC */

/* Exported macro ------------------------------------------------------------*/
/* USER CODE BEGIN EM */

/* USER CODE END EM */

void HAL_TIM_MspPostInit(TIM_HandleTypeDef *htim);

/* Exported functions prototypes ---------------------------------------------*/
void Error_Handler(void);

/* USER CODE BEGIN EFP */

/* USER CODE END EFP */

/* Private defines -----------------------------------------------------------*/
#define Onboard_Button_Pin GPIO_PIN_13
#define Onboard_Button_GPIO_Port GPIOC
#define VOUTB_Pin GPIO_PIN_3
#define VOUTB_GPIO_Port GPIOF
#define V_CS_Pin GPIO_PIN_0
#define V_CS_GPIO_Port GPIOC
#define VOUTA_Pin GPIO_PIN_3
#define VOUTA_GPIO_Port GPIOC
#define Motor_ENA_Pin GPIO_PIN_15
#define Motor_ENA_GPIO_Port GPIOF
#define Motor_ENB_Pin GPIO_PIN_13
#define Motor_ENB_GPIO_Port GPIOE
#define Motor_INA_Pin GPIO_PIN_9
#define Motor_INA_GPIO_Port GPIOG
#define Motor_INB_Pin GPIO_PIN_14
#define Motor_INB_GPIO_Port GPIOG
#define Onboard_LED_Pin GPIO_PIN_7
#define Onboard_LED_GPIO_Port GPIOB

/* USER CODE BEGIN Private defines */
void sample_callback(void);
void send_data(void);
void motor_set(void);
void motor_fwd(void);
void motor_rev(void);
/* USER CODE END Private defines */

#ifdef __cplusplus
}
#endif

#endif /* __MAIN_H */
